// © 2022, CGVR (https://cgvr.informatik.uni-bremen.de/),
// Author: Andre Mühlenbrock (muehlenb@uni-bremen.de)

// Include the GUI:
#include <imgui.h>
#include <backends/imgui_impl_glfw.h>
#include <backends/imgui_impl_opengl3.h>

// Include functions to display Mat4f and Vec4f in the GUI:
# include <imgui_cg1_helpers.h>

// Include OpenGL3.3 Core functions:
#include <glad/glad.h>

// Include GLFW/OpenGL:
#include <GLFW/glfw3.h>

// Include IO utilities:
#include <iostream>

// Time utils:
#include <chrono>

// Include Vec4f class:
#include "Vec4f.h"

// Include Mat4f class:
#include "Mat4f.h"

// Include our shader wrapper class:
#include "Shader.h"

// Include our mesh class:
#include "Mesh.h"

// Include our image texture class:
#include "Texture2D.h"

// Include class to draw debug lines
#include "LineDrawer.h"

// Stores the time when the program was started:
auto programStartTime = std::chrono::high_resolution_clock::now();

/**
 * This function generates and returns a cube mesh.
 */
Mesh generateCube() {
    // Create triangles for a cube:
    std::vector<Triangle> cubeTriangles;
    {
        // Vertex data for face 0 (front):
        cubeTriangles.push_back(Triangle(
            Vertex(Vec4f(0.0f, 0.0f, 0.0f), Vec4f(0.0f, 0.0f, 1.0f, 0.0f)),
            Vertex(Vec4f(1.0f, 0.0f, 0.0f), Vec4f(0.0f, 0.0f, 1.0f, 0.0f)),
            Vertex(Vec4f(1.0f, 1.0f, 0.0f), Vec4f(0.0f, 0.0f, 1.0f, 0.0f))

        ));
        cubeTriangles.push_back(Triangle(
            Vertex(Vec4f(0.0f, 0.0f, 0.0f), Vec4f(0.0f, 0.0f, 1.0f, 0.0f)),
            Vertex(Vec4f(1.0f, 1.0f, 0.0f), Vec4f(0.0f, 0.0f, 1.0f, 0.0f)),
            Vertex(Vec4f(0.0f, 1.0f, 0.0f), Vec4f(0.0f, 0.0f, 1.0f, 0.0f))

        ));

        // Vertex data for face 1 (right):
        cubeTriangles.push_back(Triangle(
            Vertex(Vec4f(1.0f, 0.0f, 0.0f), Vec4f(1.0f, 0.0f, 0.0f, 0.0f)),
            Vertex(Vec4f(1.0f, 0.0f, -1.0f), Vec4f(1.0f, 0.0f, 0.0f, 0.0f)),
            Vertex(Vec4f(1.0f, 1.0f, -1.0f), Vec4f(1.0f, 0.0f, 0.0f, 0.0f))

        ));
        cubeTriangles.push_back(Triangle(
            Vertex(Vec4f(1.0f, 0.0f, 0.0f), Vec4f(1.0f, 0.0f, 0.0f, 0.0f)),
            Vertex(Vec4f(1.0f, 1.0f, -1.0f), Vec4f(1.0f, 0.0f, 0.0f, 0.0f)),
            Vertex(Vec4f(1.0f, 1.0f, 0.0f), Vec4f(1.0f, 0.0f, 0.0f, 0.0f))

        ));

        // Vertex data for face 2 (back):
        cubeTriangles.push_back(Triangle(
            Vertex(Vec4f(0.0f, 0.0f, -1.0f), Vec4f(0.0f, 0.0f, -1.0f, 0.0f)),
            Vertex(Vec4f(1.0f, 1.0f, -1.0f), Vec4f(0.0f, 0.0f, -1.0f, 0.0f)),
            Vertex(Vec4f(1.0f, 0.0f, -1.0f), Vec4f(0.0f, 0.0f, -1.0f, 0.0f))

        ));
        cubeTriangles.push_back(Triangle(
            Vertex(Vec4f(0.0f, 0.0f, -1.0f), Vec4f(0.0f, 0.0f, -1.0f, 0.0f)),
            Vertex(Vec4f(0.0f, 1.0f, -1.0f), Vec4f(0.0f, 0.0f, -1.0f, 0.0f)),
            Vertex(Vec4f(1.0f, 1.0f, -1.0f), Vec4f(0.0f, 0.0f, -1.0f, 0.0f))

        ));

        // Vertex data for face 3 (left):
        cubeTriangles.push_back(Triangle(
            Vertex(Vec4f(0.0f, 0.0f, 0.0f), Vec4f(-1.0f, 0.0f, 0.0f, 0.0f)),
            Vertex(Vec4f(0.0f, 1.0f, -1.0f), Vec4f(-1.0f, 0.0f, 0.0f, 0.0f)),
            Vertex(Vec4f(0.0f, 0.0f, -1.0f), Vec4f(-1.0f, 0.0f, 0.0f, 0.0f))

        ));
        cubeTriangles.push_back(Triangle(
            Vertex(Vec4f(0.0f, 0.0f, 0.0f), Vec4f(-1.0f, 0.0f, 0.0f, 0.0f)),
            Vertex(Vec4f(0.0f, 1.0f, 0.0f), Vec4f(-1.0f, 0.0f, 0.0f, 0.0f)),
            Vertex(Vec4f(0.0f, 1.0f, -1.0f), Vec4f(-1.0f, 0.0f, 0.0f, 0.0f))

        ));

        // Vertex data for face 4 (bottom):
        cubeTriangles.push_back(Triangle(
            Vertex(Vec4f(0.0f, 0.0f, 0.0f), Vec4f(0.0f, -1.0f, 0.0f, 0.0f)),
            Vertex(Vec4f(1.0f, 0.0f, -1.0f), Vec4f(0.0f, -1.0f, 0.0f, 0.0f)),
            Vertex(Vec4f(1.0f, 0.0f, 0.0f), Vec4f(0.0f, -1.0f, 0.0f, 0.0f))

        ));
        cubeTriangles.push_back(Triangle(
            Vertex(Vec4f(0.0f, 0.0f, 0.0f), Vec4f(0.0f, -1.0f, 0.0f, 0.0f)),
            Vertex(Vec4f(0.0f, 0.0f, -1.0f), Vec4f(0.0f, -1.0f, 0.0f, 0.0f)),
            Vertex(Vec4f(1.0f, 0.0f, -1.0f), Vec4f(0.0f, -1.0f, 0.0f, 0.0f))

        ));

        // Vertex data for face 5 (top):
        cubeTriangles.push_back(Triangle(
            Vertex(Vec4f(0.0f, 1.0f, 0.0f), Vec4f(0.0f, 1.0f, 0.0f, 0.0f)),
            Vertex(Vec4f(1.0f, 1.0f, 0.0f), Vec4f(0.0f, 1.0f, 0.0f, 0.0f)),
            Vertex(Vec4f(1.0f, 1.0f, -1.0f), Vec4f(0.0f, 1.0f, 0.0f, 0.0f))

        ));
        cubeTriangles.push_back(Triangle(
            Vertex(Vec4f(0.0f, 1.0f, 0.0f), Vec4f(0.0f, 1.0f, 0.0f, 0.0f)),
            Vertex(Vec4f(1.0f, 1.0f, -1.0f), Vec4f(0.0f, 1.0f, 0.0f, 0.0f)),
            Vertex(Vec4f(0.0f, 1.0f, -1.0f), Vec4f(0.0f, 1.0f, 0.0f, 0.0f))

        ));
    }

    return Mesh(cubeTriangles);
}

/**
 * This function generates and returns a plane mesh to render water.
 */
Mesh generateWaterPlane(){
    // Create triangles for a plane:
    std::vector<Triangle> planeTriangles;
    {

        planeTriangles.push_back(Triangle(
            Vertex(Vec4f(-1.0f,-1.0f,0.99f), Vec4f(0.0f, 0.0f, 1.0f, 0.0f), Vec4f(), TexCoord(0,0)),
            Vertex(Vec4f( 1.0f,-1.0f,0.99f), Vec4f(0.0f, 0.0f, 1.0f, 0.0f), Vec4f(), TexCoord(1,0)),
            Vertex(Vec4f( 1.0f,-0.2f,0.99f), Vec4f(0.0f, 0.0f, 1.0f, 0.0f), Vec4f(), TexCoord(1,1))
        ));


        planeTriangles.push_back(Triangle(
            Vertex(Vec4f( 1.0f,-0.2f,0.99f), Vec4f(0.0f, 0.0f, 1.0f, 0.0f), Vec4f(), TexCoord(1,1)),
            Vertex(Vec4f(-1.0f,-0.2f,0.99f), Vec4f(0.0f, 0.0f, 1.0f, 0.0f), Vec4f(), TexCoord(0,1)),
            Vertex(Vec4f(-1.0f,-1.0f,0.99f), Vec4f(0.0f, 0.0f, 1.0f, 0.0f), Vec4f(), TexCoord(0,0))
        ));
    }

    return Mesh(planeTriangles);
}

/** TODO
* This function computes the intersection between a line (parameter form) and a plane (normal form)
* p0 = start position of the line
* d = direction vector of the line
* n = normal of the plane
* c = a point on the plane
* returns the line parameter t where the intersection occurs
*/
float intersectPlane(const Vec4f& p0, const Vec4f& d, const Vec4f& n, const Vec4f& c)
{
    float t = -1;

    //TODO Implement the intersection test between line and plane, return computed line parameter t
    return t;
}

/**
 * Initializes the application including the GUI which you can
 * use to test the Vec4f and Mat4f functionaliy.
 */

int main(int, char**)
{
    // Setup window:
    glfwSetErrorCallback([](int error, const char* description) {
        fprintf(stderr, "Glfw Error %d: %s\n", error, description);
    });

    if (!glfwInit())
        return 1;

    // Decide GL+GLSL versions:
    const char* glsl_version = "#version 330 core";
    glfwWindowHint(GLFW_CONTEXT_VERSION_MAJOR, 3);
    glfwWindowHint(GLFW_CONTEXT_VERSION_MINOR, 3);
    glfwWindowHint(GLFW_OPENGL_PROFILE, GLFW_OPENGL_CORE_PROFILE);
    glfwWindowHint(GLFW_OPENGL_FORWARD_COMPAT, GL_TRUE);

    // Create window with graphics context:
    GLFWwindow* window = glfwCreateWindow(900, 560, "Shader Framework (Computergrafik 1)", NULL, NULL);
    if (window == NULL)
        return 1;

    glfwMakeContextCurrent(window);
    glfwSwapInterval(1);

    // Initialize GLAD functions:
    if (!gladLoadGLLoader((GLADloadproc)glfwGetProcAddress))
    {
        std::cout << "Failed to initialize GLAD" << std::endl;
        return -1;
    }

    // Setup Dear ImGui context:
    IMGUI_CHECKVERSION();
    ImGui::CreateContext();
    ImGuiIO& io = ImGui::GetIO(); (void)io;
    io.ConfigFlags |= ImGuiConfigFlags_NavEnableKeyboard;     // Enable Keyboard Controls

    glEnable(GL_CULL_FACE);
    glEnable(GL_DEPTH_TEST);

    // Setup Dear ImGui style:
    ImGui::StyleColorsDark();

    // Setup Platform/Renderer backends:
    ImGui_ImplGlfw_InitForOpenGL(window, true);
    ImGui_ImplOpenGL3_Init(glsl_version);

    // Load Roboto Font:
    io.Fonts->AddFontFromFileTTF(CMAKE_SOURCE_DIR "/lib/imgui-1.88/misc/fonts/Roboto-Medium.ttf" , 16.0f);

    // Load the shader for the cube:
    Shader cubeShader(
        CMAKE_SOURCE_DIR "/shader/cube.vert",
        CMAKE_SOURCE_DIR "/shader/cube.frag"
    );


    // Load the shader for the debug line:
    Shader lineDrawerShader(
        CMAKE_SOURCE_DIR "/shader/lineDrawer.vert",
        CMAKE_SOURCE_DIR "/shader/lineDrawer.frag"
    );

    // Gerenates a plane for water:
    Mesh plane = generateWaterPlane();

	// generates a cube to render our line
    Mesh cube = generateCube();

    // Load the shader for the plane:
    Shader planeShader(
        CMAKE_SOURCE_DIR "/shader/plane.vert",
        CMAKE_SOURCE_DIR "/shader/plane.frag"
    );

    Texture2D waterTexture(CMAKE_SOURCE_DIR "/textures/water.jpg");

    // Colors for the algorithm:
    Vec4f startColor(1, 0, 0);
    Vec4f endColor(0, 1, 0);

    // Camera Mode (0: XY-Plane, 1: XZ-Plane, 2: ZY-Plane)
    int camera_mode = 0;
    const char* camera_mode_labels[] = { "XY-Ebene", "XZ-Ebene", "ZY-Ebene" };
    Mat4f modelMatrix;

    // Main loop which is executed every frame until the window is closed:
    while (!glfwWindowShouldClose(window))
    {
        // Processes all glfw events:
        glfwPollEvents();

        // Calculate elapsed time since program was started:
        float runTime = std::chrono::duration_cast<std::chrono::milliseconds>(std::chrono::high_resolution_clock::now() - programStartTime).count() / 1000.f;

        // Get the size of the window:
        int display_w, display_h;
        glfwGetFramebufferSize(window, &display_w, &display_h);

        // Create the GUI:
        {
            // Start the Dear ImGui frame
            ImGui_ImplOpenGL3_NewFrame();
            ImGui_ImplGlfw_NewFrame();
            ImGui::NewFrame();

            // Create Window with settings:
            ImGui::SetNextWindowPos(ImVec2(0,0));
            ImGui::SetNextWindowSize(ImVec2(float(300),float(display_h)));
            ImGui::Begin("Einstellungen", nullptr, ImGuiWindowFlags_NoMove | ImGuiWindowFlags_NoCollapse | ImGuiWindowFlags_NoResize);
            ImGui::ColorEdit3("startColor", &startColor.x, ImGuiColorEditFlags_Float);
            ImGui::Separator();
            ImGui::ColorEdit3("endColor", &endColor.x, ImGuiColorEditFlags_Float);
            ImGui::Separator();
            ImGui::Combo("Ansicht", &camera_mode, camera_mode_labels, 3);
            ImGui::End();

            // Generate a model matrix so that the object can be viewed from the preferred
            // side:
            {
                // Depending on the chosen view, change the modelMatrix:
                std::fill(std::begin(modelMatrix.data), std::end(modelMatrix.data), 0.0f);
                modelMatrix.data[15] = 1;
                if(camera_mode == 0)
                    modelMatrix.data[0] = modelMatrix.data[5] = modelMatrix.data[10] = 1;
                else if(camera_mode == 1)
                    modelMatrix.data[9] = modelMatrix.data[0] = -(modelMatrix.data[6] = -1);
                else if(camera_mode == 2)
                    modelMatrix.data[2] = -(modelMatrix.data[8] = modelMatrix.data[5] = 1);
            }
        }

        // Setup the GL viewport
        glViewport(300, 0, display_w - 300, display_h);
        glClearColor(0.6f, 0.725f, 0.8f, 1.0f);
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

        // Render plane:
        {
            // Activate our shader:
            planeShader.bind();

            // Bind our water texture to texture slot 0:
            waterTexture.bind(0);

            // Tell our shader that the sampler of 'waterImage' should use texture slot 0:
            planeShader.setUniform("waterImage", 0);

            // Upload the value of the runTime-variable to the time-variable in the shader:
            planeShader.setUniform("time", runTime);

            // Draw our triangle:
            plane.render(false);
        }


        // Render line out of cubes: TODO 3.a
        {
            // Start and end of our line:
            Vec4f start{ 0,0,0,10 };
            Vec4f end{ 7,2,-4 ,10};

            // Activate our debug-line shader:
            lineDrawerShader.bind();

			// Create our debug-line drawer object and draw a target line
            LineDrawer drawer;
            drawer.draw(start, end, Vec4f(0, 1, 1), modelMatrix);

            // Activate our cube shader:
            cubeShader.bind();
			
			// Render first cube at start of our line
            cubeShader.setUniform("position", start);
            cubeShader.setUniform("color", Vec4f(0.05, 0, 0));
			cubeShader.setUniform("model", modelMatrix);
            cube.render(true);

            Vec4f current = start;
			
			// TODO Render the rest of the line out of cubes using the rasterization algorithm from the lecture
        }

        // Render colored line: TODO 3b
        {
            // Activate our shader:
            cubeShader.bind();

			// The positions of our individual line segments
            std::vector<Vec4f> positions = { Vec4f(-3,-10,0), Vec4f(-3,-9,0), Vec4f(-2,-8,0), Vec4f(-2,-7,0), Vec4f(-2,-6,0), Vec4f(-2,-5,0)
                            , Vec4f(-1,-4,0), Vec4f(-1,-3,0), Vec4f(-1,-2,0), Vec4f(0,-1,0), Vec4f(0,0,0), Vec4f(0,1,0),
                            Vec4f(1,2,0), Vec4f(1,3,0), Vec4f(1,4,0), Vec4f(1,5,0), Vec4f(2,6,0), Vec4f(2,7,0), Vec4f(2,8,0), Vec4f(3,9,0), Vec4f(3,10,0) };
            for (int i=0; i<positions.size();++i)
            {
                positions[i] = positions[i] + Vec4f(-3, 0, 4,10);
            }
			
			// TODO Implement the incremental color interpolation from the lecture to color the line using the variables startColor and endColor
            // Differenzvariablen initialisieren

            float a = 0;
            for (int h = 0; h<positions.size(); ++h)
            {
                // Neue Farbe durch Addition
                Vec4f c(1,1,1); // Interpolierte Farbe
                
                // Gamma correction. Has to be applied after the color was computed and before it is used. Don't change it
                static float gamma = 1.8f;
                c = Vec4f(pow(c.x, 1.0/gamma), pow(c.y, 1.0/gamma), pow(c.z, 1.0/gamma));

				// Render our line
                cubeShader.setUniform("position", positions[h]);
                cubeShader.setUniform("color", c);
                cubeShader.setUniform("model", modelMatrix);
                cube.render(false);
            }
        }

        // Render the GUI and draw it to the screen:
        ImGui::Render();
        ImGui_ImplOpenGL3_RenderDrawData(ImGui::GetDrawData());

        // Swap Buffers:
        glfwSwapBuffers(window);
    }

    // Cleanup
    ImGui_ImplOpenGL3_Shutdown();
    ImGui_ImplGlfw_Shutdown();
    ImGui::DestroyContext();

    glfwDestroyWindow(window);
    glfwTerminate();

    return 0;
}

